/*!
 * \file ma_sim_model.v
 * \author Benjamin Zhai (benjamin.zhai@sensima.com)
 * \version 0.2
 * \date 20131101
 * \brief MA sensor behavioural simulation model
 *
 * Copyright © 2013 Sensima Technology SA \n
 * Company Confidential \n
 *
 * Project: MA-FTDI bridge \n
 * Module: Interface and buffer between MA sensor and FTDI \n
 *
 * Rivision History: \n
 * 0.1     - rev 0     - 20130930 - Initial version \n
 * 0.2     - rev       - 20131101 - Bug fixes \n
 *                                  Bidirectional tranmission \n
 *                                  Timing violations fixed \n
 *                                  Added internal registers and flashing \n
 * \todo  
 *        - 
 */

`timescale 1ns / 100ps

module ma_sim_model (
        input  MA_SCLK_i,
        input  MA_MOSI_i,
        output MA_MISO_o,
        input  MA_SS_n_i,
        output MA_DRDY_o
        );

reg CS_n; //filtered chip select
integer old_time;
integer clock_cntr;
reg [0:11] dev_id;
reg MA_DRDY_reg,MA_MISO_reg;
reg [23:0] data_in;

initial
begin
    CS_n = 1'bx;
    clock_cntr = 0;
    old_time = -255;
    dev_id = 12'h25a; //constant
    MA_DRDY_reg = 1'b0; //data ready
    #25 MA_DRDY_reg = 1'b1;
end

// data ready control
always@(negedge MA_SS_n_i)
begin
    MA_DRDY_reg = 1'b0;
    #25 CS_n = 1'b0;//t_csL
    if (!MA_SCLK_i)
        $display($time, ": WARNING: t_csL not respected. SCLK edge ignored.");
    #(1818:1887:2000) MA_DRDY_reg = 1'b1;  // data ready in 2us
end

// t_idle timing violation check
always@(posedge MA_SS_n_i)
begin
    CS_n = 1'b1;
    clock_cntr = 0;
    #20 //t_idle
    if (!MA_SS_n_i)
        $display($time, ": WARNING: t_idle not respected. Data is corrupted.");
end

// write to register
always@(posedge MA_SCLK_i)
begin
    if (!CS_n)
    begin
        clock_cntr = clock_cntr+1;
        data_in = (data_in << 1) & MA_MOSI_i;
		
    end
    #(3:10:25) //t_csH
    if (MA_SS_n_i)
        $display($time, ": WARNING: t_csH not respected. Data written is corrupted.");
end

// update data
always@(negedge MA_SCLK_i)
begin
    if (clock_cntr <= 7) // sensor data
    begin
        #(5:10:15) MA_MISO_reg = {$random(($time*609757)%105983)} % 2; //random sensor data
    end
	if ((clock_cntr > 7) && (clock_cntr <= 15)) // sensor data or register
    begin
		if (data_in[23:20] == 4'b0001) // command is read
			#(5:10:15) MA_MISO_reg = 0; // assuming 0 in all registers
        else
			#(5:10:15) MA_MISO_reg = {$random(($time*609757)%105983)} % 2; //random sensor data
    end
    if ((clock_cntr > 15) && (clock_cntr <= 23)) // counter
    begin
        #(5:10:15) MA_MISO_reg = 1; // -> Counter
    end
    if ((clock_cntr > 23) && (clock_cntr <= 35)) // increment
    begin
        #(5:10:15) MA_MISO_reg = 1;
    end
    if ((clock_cntr > 35) && (clock_cntr <= 47)) // ID
    begin
        #(5:10:15) MA_MISO_reg = dev_id[clock_cntr-36]; 
    end
end

// identifies clock speed
always@(MA_SCLK_i)
begin
    if ($time-old_time < 20)
        $display($time, ": WARNING: t_sclk not respected. SCLK clock rate too fast."); // t_min = 40 ns, f_max=25MHz
    old_time = $time;
end

assign MA_DRDY_o = MA_DRDY_reg;
assign MA_MISO_o = MA_MISO_reg;

endmodule
