/*!
 * \file fifo.v
 * \author Benjamin Zhai (benjamin.zhai@sensima.com)
 * \version 0.2
 * \date 20131101
 * \brief Generic FIFO core
 *
 * Copyright © 2013 Sensima Technology SA \n
 * Company Confidential \n
 *
 * Project: MA-FTDI bridge \n
 * Module: Interface and buffer between MA sensor and FTDI \n
 *
 * Rivision History: \n
 * 0.1     - rev 0     - 20130930 - Initial version \n
 * 0.2     - rev 46    - 20131104 - Bug fixes \n
 *                                  Modifled to use SPI channel 0 to FTDI \n
 * 0.3     - rev       -          - Bidirectional tranmission \n
 * 0.4     - rev       - 20131107 - Modified empty and full behaviour \n
 * \todo  
 *        - 
 */



module fifo (
    clk_i,
    wren_i,
    rden_i,
    data_i,
    data_o,
    empty_o,
    full_o,
    rst_i,
    ovf_flush_i,
    // Ports below used for debug only
    data_cntr_o
    );

// FIFO constants
parameter DATA_WIDTH = 24;
parameter ADDR_WIDTH = 11;
parameter DEPTH = (1 << ADDR_WIDTH);

// Port declarations
input                   clk_i;
input                   rst_i;
input                   wren_i;
input                   rden_i;
input  [DATA_WIDTH-1:0] data_i;
input                   ovf_flush_i;
output [DATA_WIDTH-1:0] data_o;
output                  empty_o;
output                  full_o;
output [DATA_WIDTH-1:0] data_cntr_o;

// Internal Registers
reg  [DATA_WIDTH-1:0] memory [0:DEPTH-1];
reg  [ADDR_WIDTH-1:0] write_ptr; 
reg  [ADDR_WIDTH-1:0] read_ptr; 
reg  [DATA_WIDTH-1:0] data_reg;
reg  [ADDR_WIDTH-1:0] data_cntr;
wire full_o;
wire empty_o;

// Combinatorial Logic
// FIFO full and empty logic

assign empty_o = (data_cntr == 0);
assign full_o  = (data_cntr == (DEPTH-1));
assign data_o  = data_reg;
assign data_cntr_o = data_cntr;

// Sequential Logic
// Read & Write
always@(posedge clk_i)
begin
    if ((wren_i == 1'b1) && (rden_i == 1'b1)) // write/read conflict check -> write and increment write pointer & read and increment read pointer
	begin
		memory[write_ptr] <= data_i;
        write_ptr <= write_ptr + 1;
		data_reg <= memory[read_ptr];
        read_ptr <= read_ptr + 1;
	end
	else if ((wren_i == 1'b1) && (full_o == 1'b0)) // write enabled and not full -> write and increment write pointer 
    begin
        memory[write_ptr] <= data_i;
        write_ptr <= write_ptr + 1;
    end
    else if ((rden_i == 1'b1) && (empty_o == 1'b0)) // read enabled and not empty -> read and increment read pointer
    begin
        data_reg <= memory[read_ptr];
        read_ptr <= read_ptr + 1;
    end
    else if ((wren_i == 1'b1) && (full_o == 1'b1) && (ovf_flush_i == 1'b1)) // write enabled and full and in flush mode -> write and increment read & write pointer 
    begin
        memory[write_ptr] <= data_i;
        write_ptr <= write_ptr + 1;
        read_ptr <= read_ptr +1;
    end
    else if ((rden_i == 1'b1) && (empty_o == 1'b1)) // read enabled and empty -> output 0 (and set empty flag)
    begin
        data_reg <= 0;
    end

    // write enabled and full and not in flush mode -> do nothing, data is discarded
    if (rst_i)
    begin
        write_ptr <= 0;
        read_ptr  <= 0;
        data_reg <= 0;
    end
end

// SYNTHESISABLE: initialise memory with test data
// initial $readmemh("meminit.txt", memory);

// Status counter for empty and full flags
always@(posedge clk_i)
begin
    if (rst_i)
        data_cntr <= 0;
    else if ((rden_i == 1'b1) && (wren_i == 1'b0) && (data_cntr != 0)) //Read but no write
        data_cntr <= data_cntr - 1; // Decrement data counter
    else if ((rden_i == 1'b0) && (wren_i == 1'b1) && (data_cntr != DEPTH-1)) //Write but no read
        data_cntr <= data_cntr + 1; // Increment data counter
    // Nothing changes when read and write
end
endmodule 
