/*! 
 * \file ftdibridge.v
 * \author Benjamin Zhai (benjamin.zhai@sensima.com)
 * \version 0.2
 * \date 20131101
 * \brief Interface and buffer between MA sensor and FTDI
 *
 * Copyright © 2013 Sensima Technology SA \n
 * Company Confidential \n
 *
 * Project: MA-FTDI bridge \n
 * Module: Interface and buffer between MA sensor and FTDI \n
 *
 * Rivision History: \n
 * 0.1     - rev 0     - 20130930 - Initial version \n
 * 0.2     - rev 46    - 20131104 - Bug fixes \n
 *                                  Modified to use SPI channel 0 to FTDI \n
 * 0.3     - rev 66    - 20131112 - Bug fixes - timing violation \n
 * 0.4     - rev 80    - 20131115 - New fsm\n
 * 0.5     - rev 104   - 20131206 - Changed to use Pmod connector \n
 *                                - New GB constraints\n
 * 0.6     - rev 112   - 20131209 - Various bug fixes\n
 * 0.7     - rev 119   - 20131209 - Recovered from fatal error \n
 * \todo  
 *        - test under different clock rates/divider settings
 */

`timescale 1ns / 100ps
// synthesis loop_limit 4000

module FTDI_bridge(
    input  clk_i,
    input  FTDI_SCLK_i,
    input  FTDI_MOSI_i,
    output FTDI_MISO_o,
    input  FTDI_SS_n_i,
    output MA_SCLK_o,
    output MA_MOSI_o,
    input  MA_MISO_i,
    output MA_SS_n_o,
    input  MA_DRDY_i,
    output LED1_o,
    output LED2_o,
    output LED3_o,
    output LED4_o,
    output LED5_o,
    output PWR_o,
	output GND_o
);

// Internal connections and registers
    reg  [23:0] cntr1=0;        // counter for clkdiv
    reg  [11:0] cntr2=0;        // counter for init
    wire        clk_pll;        // PLL clock output
    wire        PLL_LOCK_net;
    reg  [23:0] PLL_LOCK_buf;
    wire        gclk;           // PLL global clock output
    wire        SPI_S_dreq;
    reg         SPI_S_dreq_old;
	reg  [3:0]  SPI_S_dreq_delay;
    wire        SPI_S_wren;
    wire        SPI_S_do_valid; // two clock cycle pulse
    reg         SPI_S_do_valid_old;
    wire [23:0] SPI_S_data_in;
    wire        SPI_M_wren;     // Enable SPI master write
    wire [23:0] SPI_M_data_in;
    wire [23:0] SPI_M_do;
    wire        SPI_M_do_valid; // two clock cycle pulse
    reg         SPI_M_do_valid_old;
    wire        PGEN_M_out;
    wire        PGEN_S_out;
    wire [23:0] FIFO_data_in;
    wire [23:0] FIFO_data_out;
    wire        FIFO_wren;      // one clock cycle pulse
    wire        FIFO_rden;
    wire [23:0] FIFO_bypass;    // FIFO bypass
    wire [23:0] cmd_fifo_data_out;
    wire [23:0] cmd_fifo_data_in;
    wire        cmd_fifo_empty;
    wire        cmd_fifo_full;
    wire        cmd_fifo_wren;
    wire        cmd_fifo_rden;
    wire        MA_SCLK;       // monitor SCLK
    reg         MA_SCLK_old;
    reg  [4:0]  MA_SCLK_cntr;  // SCLK counter
    reg  [15:0] MA_timeout_cntr;
    reg  [4:0]  MA_delay;
    wire        mux1_sel;
    wire        mux2_sel;
    wire        demux1_sel;
    wire        grst;           // global reset
    reg         grst_n;
	wire		fifo_empty_net;
	wire		fifo_full_net;
    
    integer i,j;
    
// parameters (constants)
    parameter spi_clkdiv2 = 4;

/*/ Clock generation for simulation
reg PLL_LOCK_reg; 
initial
begin
    PLL_LOCK_reg = 1'b0;
    #1000  PLL_LOCK_reg = 1'b1;
end
assign PLL_LOCK_net = PLL_LOCK_reg;
assign clk_pll=clk_i;  
// Clock generation for simulation */


// PLL instantiation
    pll_mclk pll_mclk_inst (.REFERENCECLK(clk_i),
                            .PLLOUTCORE(clk_pll),
                            .PLLOUTGLOBAL(gclk),
                            .RESET(1'b1),  // Active LOW reset
                            .LOCK(PLL_LOCK_net)
                            );
// PLL instantiation */

// Global Reset generation
    always@(posedge clk_i)
    begin: grst_gen
        PLL_LOCK_buf[0] <= PLL_LOCK_net;
        for (i=23; i>0; i = i-1) // DFF chain
            PLL_LOCK_buf[i] <= PLL_LOCK_buf[i-1];
        grst_n <= PLL_LOCK_buf[23];
    end
    assign grst = ~grst_n;


// SPI slave instantiation
    spi_slave #(24, 2'b10, 2'b10, 2 ) 
    ftdi_spi_slave (.clk_i(clk_pll),
                    .spi_ssel_i(FTDI_SS_n_i),
                    .spi_sck_i(FTDI_SCLK_i),
                    .spi_mosi_i(FTDI_MOSI_i),
                    .spi_miso_o(FTDI_MISO_o),
                    .di_req_o(SPI_S_dreq),
                    .di_i(SPI_S_data_in),
                    .wren_i(SPI_S_wren),
                    .wr_ack_o(),
                    .do_valid_o(SPI_S_do_valid),
                    .do_o(cmd_fifo_data_in),
                    // debug ports below
                    .do_transfer_o(),
                    .wren_o(),
                    .rx_bit_next_o(),
                    .state_dbg_o(),
                    .sh_reg_dbg_o()
                    );

// SPI master instantiation
    spi_master #(24, 2'b11, 2'b11, 3, spi_clkdiv2) 
    fas_spi_master (.sclk_i(clk_pll),
                    .pclk_i(clk_pll),
                    .rst_i(grst),
                    .spi_ssel_o(), // not connected
                    .spi_sck_o(MA_SCLK),
                    .spi_mosi_o(MA_MOSI_o),
                    .spi_miso_i(MA_MISO_i),
                    .di_req_o(),
                    .di_i(SPI_M_data_in),
                    .wren_i(SPI_M_wren),
                    .wr_ack_o(),
                    .do_valid_o(SPI_M_do_valid),
                    .do_o(SPI_M_do),
                    // debug ports below
                    .sck_ena_o(), 
                    .sck_ena_ce_o(),
                    .do_transfer_o(),
                    .wren_o(),
                    .rx_bit_reg_o(),
                    .state_dbg_o(),
                    .core_clk_o(),
                    .core_n_clk_o(),
                    .core_ce_o(),
                    .core_n_ce_o(),
                    .sh_reg_dbg_o()
                    );

// DATA FIFO instantiation
    //defparam fifo_mem1.DATA_WIDTH = 24;
    //defparam fifo_mem1.ADDR_WIDTH = 11; // 3.5ms of data @ 500kHz = 1750 data points = 11 bit address
    fifo fifo_mem1 (.clk_i(clk_pll),
                    .wren_i(FIFO_wren),
                    .rden_i(FIFO_rden),
                    .data_i(FIFO_data_in),
                    .data_o(FIFO_data_out),
                    .empty_o(fifo_empty_net),  // FIFO empty indicator
                    .full_o(fifo_full_net),   // FIFO full indicator
                    .rst_i(grst),
                    .ovf_flush_i(1'b1), // flush unread data
                    // debug ports below
                    .data_cntr_o()
                    );
                    
// Command FIFO instantiation
    defparam cmd_fifo.DATA_WIDTH = 24;
    defparam cmd_fifo.ADDR_WIDTH = 4; // 16 commands
    fifo cmd_fifo  (.clk_i(clk_pll),
                    .wren_i(cmd_fifo_wren),
                    .rden_i(cmd_fifo_rden),
                    .data_i(cmd_fifo_data_in),
                    .data_o(cmd_fifo_data_out),
                    .empty_o(cmd_fifo_empty), 
                    .full_o(cmd_fifo_full),  
                    .rst_i(grst),
                    .ovf_flush_i(1'b0), // do not flush commands, discard excess
                    // debug ports below
                    .data_cntr_o()
                    );
                    
// MUX and DEMUX instantiation
defparam mux1.WIDTH=24;
mux mux1 (  .en_i(1'b1),
            .i0_i(FIFO_data_out),
            .i1_i({FIFO_bypass[23:16],FIFO_bypass[14:0],1'b0}),  // excessive 1'b0 on bit 15 
            .sel_i(mux1_sel),
            .dout_o(SPI_S_data_in)
        );

defparam mux2.WIDTH=24;
mux mux2 (  .en_i(1'b1),
            .i0_i(24'b0),
            .i1_i(cmd_fifo_data_out),
            .sel_i(mux2_sel),
            .dout_o(SPI_M_data_in)
        );

defparam mux3.WIDTH=1;
mux mux3 (  .en_i(1'b1),
            .i0_i(PGEN_S_out),
            .i1_i(1'b0),
            .sel_i((cmd_fifo_data_in==24'd0)||(cmd_fifo_data_in==24'hffffff)),
            .dout_o(cmd_fifo_wren)
        );
        
defparam demux1.WIDTH=24;
demux demux1 (  .en_i(1'b1),
                .o0_o(FIFO_data_in),
                .o1_o(FIFO_bypass),
                .sel_i(demux1_sel),
                .din_i(SPI_M_do)
            );

defparam demux2.WIDTH=1;
demux demux2 (  .en_i(1'b1),
                .o0_o(FIFO_wren),
                .o1_o(), // NC
                .sel_i(demux1_sel), // connected to demux1_sel
                .din_i(PGEN_M_out)
            );

// Pulse generator glue logic for FIFO WREN
// Converts multi clock cycle pulse to single clock cycle pulse
    always@(posedge clk_pll)
    begin: pgen_m
        if (grst)
            SPI_M_do_valid_old <= 1'b0;
        else
            SPI_M_do_valid_old <= SPI_M_do_valid;
    end
    assign PGEN_M_out = SPI_M_do_valid & (~SPI_M_do_valid_old);
    
// Pulse generator glue logic for FIFO WREN
// Converts multi clock cycle pulse to single clock cycle pulse
    always@(posedge clk_pll)
    begin: pgen_s
        if (grst)
            SPI_S_do_valid_old <= 1'b0;
        else
            SPI_S_do_valid_old <= SPI_S_do_valid;
    end
    assign PGEN_S_out = SPI_S_do_valid & (~SPI_S_do_valid_old);
	
    
// MA fsm: fsm2
    parameter [3:0] // synopsis enum code
        MA_IDLE     = 4'd0,
        MA_RDY      = 4'd1,
        MA_XFER     = 4'd2,
        MA_RSVD     = 4'd3,
        MA_CMD_RD   = 4'd4,
        MA_WRITE    = 4'd5,
        MA_READ     = 4'd6,
		MA_WAIT		= 4'd7,
		MA_W_XFER   = 4'd8,
		MA_R_XFER   = 4'd9;
    // synopsis state_vector ma_state
    reg [3:0] // synopsis enum code
        ma_state,ma_next;
    always@(*)
    begin: fsm2
        ma_next = 4'bx;
        case(ma_state)
            MA_IDLE: begin
                if (cmd_fifo_data_out[23:20] == 4'b1101) // reserved
                    ma_next = MA_RSVD;
                else if (cmd_fifo_data_out[23:20] == 4'b0010) // write
                    ma_next = MA_WRITE;
                else if (cmd_fifo_data_out[23:20] == 4'b0001) // read
                    ma_next = MA_READ;
                else if ( (cmd_fifo_empty == 1'b0) || (cmd_fifo_data_out != 24'b0) )
                    ma_next = MA_CMD_RD;
                else if (MA_DRDY_i)
                    ma_next = MA_RDY;
                else
                    ma_next = MA_IDLE;
            end
            MA_RDY: begin
                if (MA_delay > 5'b10000)
                    ma_next = MA_XFER;
                else
                    ma_next = MA_RDY;
            end
            MA_XFER: begin
                if ((MA_SCLK_cntr == 5'd24)||(MA_timeout_cntr > 16'h8000))
                    ma_next = MA_IDLE;
                else
                    ma_next = MA_XFER;
            end
            MA_RSVD: begin
                ma_next = MA_IDLE;
            end
            MA_CMD_RD: begin
                ma_next = MA_IDLE;
            end
            MA_WRITE: begin
                if (MA_delay > 5'b10000)
                    ma_next = MA_W_XFER;
                else
                    ma_next = MA_WRITE;
            end
            MA_READ: begin
                if (MA_delay > 5'b10000)
                    ma_next = MA_R_XFER;
                else
                    ma_next = MA_READ;
            end
			MA_WAIT: begin
				if (SPI_S_dreq_delay[3] == 1'b1)  // wait until data is fetched by SPI slave
					ma_next = MA_IDLE;
				else
					ma_next = MA_WAIT;
            end
			MA_W_XFER: begin
				if ((MA_SCLK_cntr == 5'd24)||(MA_timeout_cntr > 16'h8000))
                    ma_next = MA_IDLE;
                else
                    ma_next = MA_W_XFER;
			end
			MA_R_XFER: begin
				if ((MA_SCLK_cntr == 5'd24)||(MA_timeout_cntr > 16'h8000))
                    ma_next = MA_WAIT;
                else
                    ma_next = MA_R_XFER;
			end
			default: begin
				ma_next = MA_IDLE;
			end
        endcase
    end
    always@(posedge clk_pll)
    begin: fsm2_next
        if (grst)
        begin
            ma_state <= MA_IDLE;
            MA_SCLK_old <= 1'b1;
        end
        else begin
            ma_state <= ma_next;
            MA_SCLK_old <= MA_SCLK;
        end
    end
    always@(posedge clk_pll) // MA delay 
    begin: fsm2_MA_delay
        if ( (ma_state == MA_RDY) || (ma_state == MA_WRITE) || (ma_state == MA_READ) )
            MA_delay <= MA_delay + 1;
        else
            MA_delay <= 5'b0;
    end
	always@(posedge clk_pll) // WAIT delay 
    begin: fsm2_WAIT_delay
		if (grst)
		begin
			for (j=0; j<=3; j=j+1)
				SPI_S_dreq_delay[j] <= 0;
		end
		else begin
			SPI_S_dreq_delay[0] <= SPI_S_dreq;
			for (j=1; j<=3; j=j+1)
				SPI_S_dreq_delay[j] <= SPI_S_dreq_delay[j-1];
		end
    end
    always@(posedge clk_pll) // MA_XFER counter/timeout
    begin: fsm2_cntr_timeout
        if ( (ma_state == MA_XFER) || (ma_state == MA_W_XFER) || (ma_state == MA_R_XFER) )
        begin
            MA_timeout_cntr <= MA_timeout_cntr + 1;
            if ( (MA_SCLK_old == 1'b0) && (MA_SCLK == 1'b1) ) // rising edge
                MA_SCLK_cntr <= MA_SCLK_cntr + 1;
            else 
                MA_SCLK_cntr <= MA_SCLK_cntr;
        end
        else begin
            MA_SCLK_cntr <= 5'b0;
            MA_timeout_cntr <= 16'b0;
        end
    end
    assign SPI_M_wren = ( ((ma_state == MA_RDY)&&(MA_delay > 5'b10000)) || ((ma_state == MA_READ)&&(MA_delay > 5'b10000)) || ((ma_state == MA_WRITE)&&(MA_delay > 5'b10000)) )?1'b1:1'b0;
    assign MA_SS_n_o = ((ma_state == MA_READ)||(ma_state == MA_R_XFER)||(ma_state == MA_WRITE)||(ma_state == MA_W_XFER)||(ma_state == MA_RDY)||(ma_state == MA_XFER))?1'b0:1'b1;
    assign cmd_fifo_rden = ( (ma_state == MA_CMD_RD)||((ma_state == MA_READ)&&(MA_delay > 5'b10000)) || ((ma_state == MA_WRITE)&&(MA_delay > 5'b10000))||(ma_state == MA_RSVD) )?1'b1:1'b0;
    assign demux1_sel = ( (ma_state == MA_READ) || (ma_state == MA_R_XFER) || (ma_state == MA_WAIT) )?1'b1:1'b0;
    assign mux1_sel = ( (ma_state == MA_READ) || (ma_state == MA_R_XFER) || (ma_state == MA_WAIT) )?1'b1:1'b0;
    assign mux2_sel = 1'b1;

// FIFO READ fsm: fsm1
    parameter [1:0] // synopsis enum code
        FIFO_IDLE   = 2'd0,
        FIFO_READ   = 2'd1,
        FIFO_SLEN   = 2'd2,
        FIFO_ERROR  = 2'd3;
    // synopsis state_vector state
    reg [1:0] // synopsis enum code
        fifo_state, fifo_next;
    always@(*)
    begin: fsm1
        fifo_next = 2'bx;
        case(fifo_state)
            FIFO_IDLE: begin
                if ((SPI_S_dreq_old == 1'b0)&&(SPI_S_dreq == 1'b1)) fifo_next = FIFO_READ;
                else fifo_next = FIFO_IDLE;
            end
            FIFO_READ: begin
                fifo_next = FIFO_SLEN;
            end
            FIFO_SLEN: begin
                fifo_next = FIFO_IDLE;
            end
            FIFO_ERROR: begin
                fifo_next = FIFO_IDLE;
            end
			default: begin
				fifo_next = FIFO_IDLE;
			end
        endcase
    end
    
    always@(posedge clk_pll)
    begin: fsm1_next
        if (grst)
        begin
            SPI_S_dreq_old <= 1'b0;
            fifo_state <= FIFO_IDLE;
        end
        else begin
            SPI_S_dreq_old <= SPI_S_dreq;
            fifo_state <= fifo_next;
        end
    end
    
    assign FIFO_rden = (fifo_state == FIFO_READ);
    assign SPI_S_wren = (fifo_state == FIFO_SLEN);


// Output logic
    assign LED1_o = fifo_empty_net;     // LED1 : FIFO empty indicator
    assign LED2_o = fifo_full_net;      // LED2 : FIFO full indicator
    assign LED3_o = ~MA_SS_n_o;         // LED3 : FAS read indicator
    assign LED4_o = MA_DRDY_i;          // LED4 : FAS DR indicator
    assign LED5_o = PLL_LOCK_net;          // LED5 : PLL indicator

	// assign LED1_o = ((ma_state == MA_XFER)||(ma_state == MA_IDLE)||(ma_state == MA_RDY));     // LED1 : FIFO empty indicator
    // assign LED2_o = ((ma_state == MA_RSVD)||(ma_state == MA_CMD_RD));      // LED2 : FIFO full indicator
    // assign LED3_o = (ma_state == MA_WAIT);         // LED3 : FAS read indicator
    // assign LED4_o = ((ma_state == MA_R_XFER)||(ma_state == MA_READ)||(ma_state == MA_WRITE)||(ma_state == MA_W_XFER));          // LED4 : FAS DR indicator
    // assign LED5_o = ((ma_state != MA_XFER)&&(ma_state != MA_IDLE)&&(ma_state != MA_RDY)&&(ma_state != MA_RSVD)&&(ma_state != MA_CMD_RD)&&(ma_state != MA_WAIT)&&(ma_state != MA_READ)&&(ma_state != MA_R_XFER)&&(ma_state != MA_WRITE)&&(ma_state != MA_W_XFER));          // LED5 : PLL indicator

    assign MA_SCLK_o = MA_SCLK;
    assign PWR_o = 1'b1;
	assign GND_o = 1'b0;
    
endmodule
