/*!
 * \file ftdibridge_tb.v
 * \author Benjamin Zhai (benjamin.zhai@sensima.com)
 * \version 0.2
 * \date 20131101
 * \brief Testbench of ftdibridge.v
 *
 * Copyright © 2013 Sensima Technology SA \n
 * Company Confidential \n
 *
 * Project: MA-FTDI bridge \n
 * Module: Interface and buffer between MA sensor and FTDI \n
 *
 * Rivision History: \n
 * 0.1     - rev 0     - 20130930 - Initial version \n
 * 0.2     - rev 46    - 20131104 - Bug fixes \n
 *                                  Modifled to use SPI channel 0 to FTDI \n
 * 0.3     - rev       -          - Bidirectional tranmission \n
 * \todo  
 *        - test under different clock rates/divider settings
 */
 
`timescale 1ns / 100ps
// synthesis loop_limit 4000
module FTDI_bridge_tb();

// Register instantiations
reg clk_in;
reg FTDI_SCLK_in, FTDI_MOSI_in, FTDI_SS_n_in;
integer i,j;
reg [7:0] data_reg;

wire FTDI_MISO_out;
wire MA_SCLK_net, MA_MOSI_net, MA_SS_n_net;
wire MA_MISO_net, MA_DRDY_net;
wire LED1_out, LED2_out, LED3_out, LED4_out, LED5_out; 

// DUT instantiation
FTDI_bridge DUT (   .clk_i(clk_in),
                    .FTDI_SCLK_i(FTDI_SCLK_in),
                    .FTDI_MOSI_i(FTDI_MOSI_in),
                    .FTDI_MISO_o(FTDI_MISO_out),
                    .FTDI_SS_n_i(FTDI_SS_n_in),
                    .MA_SCLK_o(MA_SCLK_net),
                    .MA_MOSI_o(MA_MOSI_net),
                    .MA_MISO_i(MA_MISO_net),
                    .MA_SS_n_o(MA_SS_n_net),
                    .MA_DRDY_i(MA_DRDY_net),
                    .LED1_o(LED1_out),
                    .LED2_o(LED2_out),
                    .LED3_o(LED3_out),
                    .LED4_o(LED4_out),
                    .LED5_o(LED5_out)
                    );

ma_sim_model ma3(
        .MA_SCLK_i(MA_SCLK_net),
        .MA_MOSI_i(MA_MOSI_net),
        .MA_MISO_o(MA_MISO_net),
        .MA_SS_n_i(MA_SS_n_net),
        .MA_DRDY_o(MA_DRDY_net)
        );

// initial block
initial
begin
    $display($time, ": ==== Starting simulation ====");
    clk_in       = 1'b0;
    FTDI_SCLK_in = 1'b0;
    FTDI_MOSI_in = 1'b1;
    FTDI_SS_n_in = 1'b1;
    @(negedge clk_in);
	#10000
	ftdi_read(12);
    #20000
    ftdi_write(24'h15aa55); // Command: READ
    #50000
    ftdi_read(9);
	#10000
    ftdi_write(24'h25aa55); // Command: Write
	#10000
	ftdi_read(9);
	#10000
    ftdi_write(24'hd5aa55); // Command: RSVD
	#10000
	ftdi_read(9);
	#10000
    ftdi_write(24'ha5aa55); // Command: None-existent
	#10000
	ftdi_read(9);
    #20000 $display($time, ": ==== Simulation complete ====");
    $stop;
end

// signal monitoring block
// initial
// begin
    // $monitor($time, ": Monitor: FTDI_SCLK_in=%b, FTDI_MOSI_in=%b, FTDI_MISO_out=%b, FTDI_SS_n_in=%b",FTDI_SCLK_in,FTDI_MOSI_in,FTDI_MISO_out,FTDI_SS_n_in);
    // $monitor($time, ": Monitor: MA_SCLK_net=%b, MA_MOSI_net=%b, MA_MISO_net=%b, MA_SS_n_net=%b, MA_DRDY_net=%b",MA_SCLK_net,MA_MOSI_net,MA_MISO_net,MA_SS_n_net,MA_DRDY_net);
// end

// clock generation
always
    forever #5 clk_in = ~clk_in;  // toggle every 5ns, 100MHz

task ftdi_read;
    input [7:0] n_o_bytes;
    begin
        FTDI_SCLK_in = 1'b0;
        FTDI_SS_n_in = 1'b1;
        FTDI_MOSI_in = 1'b1;
        @(negedge clk_in);
        FTDI_SS_n_in = 1'b0;
        #20 $display($time, ": FTDI Reading");
        for (i=0; i<n_o_bytes; i=i+1)
        begin
            for (j=7; j>=0; j=j-1)
            begin
                #25 FTDI_SCLK_in = 1'b1;
                #25 FTDI_SCLK_in = 1'b0;
                data_reg [j] = FTDI_MISO_out;
            end
            $display($time, ": Data read by FTDI: %h",data_reg);
        end
        #20 FTDI_SS_n_in = 1'b1;
    end
endtask

task ftdi_write;
    input [23:0] data_in;
    begin
        FTDI_SCLK_in = 1'b0;
        FTDI_SS_n_in = 1'b1;
        FTDI_MOSI_in = 1'b1;
        @(negedge clk_in);
        FTDI_SS_n_in = 1'b0;
        #20 $display($time, ": FTDI Writing");
        for (i=23; i>=0; i=i-1)
        begin
            FTDI_MOSI_in = data_in[i];
            #25 FTDI_SCLK_in = 1'b1;
            #25 FTDI_SCLK_in = 1'b0;
            
            $display($time, ": Data written to FTDI: %h",data_in);
        end
        #20 FTDI_SS_n_in = 1'b1;
    end
endtask

endmodule
